/*
 * $Id: da.h,v 1.1 1996/06/11 21:20:46 hjp Exp $
 *
 * Dynamic arrays -- macros for simple expansion of arrays
 *
 * Purpose
 *
 * $Log: da.h,v $
 * Revision 1.1  1996/06/11 21:20:46  hjp
 * Added efregister
 * Added Configuration for Linux-ELF.
 *
 */

#include <stdlib.h>

/*
 * What to do in case of error.
 *
 * This must be exactly one statement.
 * Applications will probably want to redefine this to something
 * less intrusive.
 */
#ifndef DA_ERROR 
#define DA_ERROR abort();
#endif

#ifndef MAX
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#endif

/* Macro: DA_GROW
 * Purpose: Return the new size of an array which had size |a| and must
 *	include index |b|.
 *	This macro may be changed by the application, but the default is
 *	expected to be useful for many applications.
 *
 * Algorithm: It is assumed that most arrays will grow linearly,
 *	that is, |b| will equal |a|. To avoid calling realloc for every
 *	element added, the size is multiplied by a constant factor.
 *	The factor is 1.5 because a factor of 2 has produced extreme
 *	fragmentation with some allocators.
 *	For the case where this expansion is not sufficient to reach
 *	index |b| we do not guess about the future growth of the array
 *	and make it just large enough.
 *	This algorithm is good enough to start at zero, but the 
 *	steps will be very small at first (0, 1, 2, 3, 4, 6, 9, 13, ...)
 *	so it might be a good idea to start with some moderate index.
 */
#ifndef DA_GROW
#define DA_GROW(a,b) MAX((a) * 3 / 2, (b) + 1)
#endif

/* Macro: DA_BASE
 * Purpose: Return the base of the dynamic array. This is useful for
 * passing to free or for duplicating a DA.
 */
#define DA_BASE(ptr) (((size_t *)(ptr))-1)
/* Macro: DA_SIZE
 * Purpose: Return the current size of the dynamic array
 */
#define DA_SIZE(ptr) (((size_t *)(ptr))[-1])

/* Macro: DA_MKIND
 * Purpose:
 *	Make sure index |ind| exists in array |ptr|.
 *	|ptr| can be a null pointer or an existing DA.
 * 	If |ptr| is a null pointer, a DA of size |ind + 1|
 *	is created. If it is a DA of <= |ind| elements
 *	it is expanded to a size determined by the DA_GROW macro.
 *	In both cases the contents of the new elements is indeterminate.
 *	Otherwise the DA is left unchanged.
 * Parameters:
 *	ptr	(inout, multi)	DA
 *	ind	(in, multi)	required index
 */
#define DA_MKIND(ptr, ind) {						\
	if (!(ptr)) {							\
		(ptr) = malloc (sizeof(size_t) +			\
				((ind) + 1) * sizeof(*(ptr)));		\
		if (!(ptr)) DA_ERROR					\
		(ptr) = (void *)((size_t *)(ptr) + 1);				\
		DA_SIZE(ptr) = (ind) + 1;				\
	} else if ((ind) >= DA_SIZE(ptr)) {				\
		size_t DA_nsiz = DA_GROW(DA_SIZE(ptr), (ind));	\
		(ptr) = realloc (DA_BASE(ptr),			\
				 sizeof(size_t) +			\
				 DA_nsiz * sizeof(*(ptr)));		\
		if (!(ptr)) DA_ERROR					\
		(ptr) = (void *)((size_t *)(ptr) + 1);				\
		DA_SIZE(ptr) = DA_nsiz;					\
	}								\
}



/* Macro: DA_MKIND
 * Purpose:
 *	Make sure index |ind| exists in array |ptr|.
 *	|ptr| can be a null pointer or an existing DA.
 * 	If |ptr| is a null pointer, a DA of size |ind + 1|
 *	is created. If it is a DA of <= |ind| elements
 *	it is expanded to a size determined by the DA_GROW macro.
 *	In both cases |null| is assigned to each of the new elements.
 *	Otherwise the DA is left unchanged.
 * Parameters:
 *	ptr	(inout, multi)	DA
 *	ind	(in, multi)	required index
 */
#define DA_MKIND_INI(ptr, ind, null) {				\
	size_t DA_i;							\
	if (!(ptr)) {							\
		(ptr) = malloc (sizeof(size_t) +			\
				((ind) + 1) * sizeof(*(ptr)));		\
		if (!(ptr)) DA_ERROR					\
		(ptr) = (void *)((size_t *)(ptr) + 1);				\
		for (DA_i = 0; DA_i <= (ind); DA_i ++) {		\
			(ptr)[DA_i] = (null);				\
		}							\
		DA_SIZE(ptr) = (ind) + 1;				\
	} else if ((ind) >= DA_SIZE(ptr)) {				\
		size_t DA_nsiz = DA_GROW(DA_SIZE(ptr), (ind));	\
		(ptr) = realloc (DA_BASE(ptr),			\
				 sizeof(size_t) +			\
				 DA_nsiz * sizeof(*(ptr)));		\
		if (!(ptr)) DA_ERROR					\
		(ptr) = (void *)((size_t *)(ptr) + 1);				\
		for (DA_i = DA_SIZE(ptr);				\
		     DA_i <= (ind);					\
		     DA_i ++) {					\
			(ptr)[DA_i] = (null);				\
		}							\
		DA_SIZE(ptr) = DA_nsiz;					\
	}								\
}
